/*
 * Line6 PODxt Pro USB driver - 0.4
 *
 * Copyright (C) 2004, 2005 Markus Grabner (grabner@icg.tu-graz.ac.at)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 */

#include <linux/config.h>

#ifdef CONFIG_USB_DEBUG
#define DEBUG 1
#endif

#include <asm/uaccess.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/usb.h>
#include <linux/time.h>

#include <sound/driver.h>
#include <sound/core.h>
#include <sound/control.h>
#include <sound/initval.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>

#include "audio.h"
#include "capture.h"
#include "driver.h"
#include "playback.h"


static int podxtpro_index[SNDRV_CARDS] = SNDRV_DEFAULT_IDX;
static char *podxtpro_id[SNDRV_CARDS] = SNDRV_DEFAULT_STR;


/* control info callback */
static int snd_podxtpro_control_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo) {
	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = 2;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 256;
	return 0;
}

/* control get callback */
static int snd_podxtpro_control_get(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
	int i;
	snd_podxtpro_audio_t *chip = snd_kcontrol_chip(kcontrol);

	for(i = 2; i--;)
		ucontrol->value.integer.value[i] = chip->volume[i];

	return 0;
}

/* control put callback */
static int snd_podxtpro_control_put(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
	int i, changed = 0;
	snd_podxtpro_audio_t *chip = snd_kcontrol_chip(kcontrol);

	for(i = 2; i--;)
		if(chip->volume[i] != ucontrol->value.integer.value[i]) {
			chip->volume[i] = ucontrol->value.integer.value[i];
			changed = 1;
		}

	return changed;
}

/* control definition */
static snd_kcontrol_new_t podxtpro_control = {
	.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	.name = "PCM Playback Volume",
	.index = 0,
	.access = SNDRV_CTL_ELEM_ACCESS_READWRITE,
	.info = snd_podxtpro_control_info,
	.get = snd_podxtpro_control_get,
	.put = snd_podxtpro_control_put
};


/* create a pcm device */
static int snd_podxtpro_new_pcm(struct snd_podxtpro_audio *chip)
{
	snd_pcm_t *pcm;
	int err;

	if((err = snd_pcm_new(chip->card, (char *)chip->podxtpro->devname, 0, 1, 1, &pcm)) < 0) 
		return err;

	pcm->private_data = chip;
	strcpy(pcm->name, chip->podxtpro->devname);
	chip->pcm = pcm;

	/* set operators */
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &snd_podxtpro_playback_ops);
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &snd_podxtpro_capture_ops);

	/* pre-allocation of buffers */
	snd_pcm_lib_preallocate_pages_for_all(pcm,
																				SNDRV_DMA_TYPE_CONTINUOUS,
																				snd_dma_continuous_data(GFP_KERNEL),
																				64 * 1024, 128 * 1024);

	return 0;
}

/* destructor */
static int snd_podxtpro_free(struct snd_podxtpro_audio *chip)
{
	return 0;
}

/* device destructor */
static int snd_podxtpro_dev_free(snd_device_t *device)
{
	struct snd_podxtpro_audio *chip = device->device_data;
	return snd_podxtpro_free(chip);
}

/*
	Initialize the PODxt Pro USB audio system.
	Create and register the sound card and mixer entries.
	Create URBs for playback and capture.
*/
int init_audio(struct usb_interface *interface)
{
	static snd_device_ops_t ops = {
		.dev_free = snd_podxtpro_dev_free,
	};

	static int dev = 0;
	snd_card_t *card;
	snd_podxtpro_audio_t *chip;
	int err;
	usb_podxtpro_t *podxtpro = usb_get_intfdata(interface);

	card = snd_card_new(podxtpro_index[dev], podxtpro_id[dev], THIS_MODULE, 0);

	if(card == NULL)
		return -ENOMEM;

	chip = kmalloc(sizeof(snd_podxtpro_audio_t), GFP_KERNEL);

	if(chip == NULL)
		return -ENOMEM;

	memset(chip, 0, sizeof(snd_podxtpro_audio_t));
	chip->card = card;
	chip->podxtpro = podxtpro;
	podxtpro->chip = chip;

	if((err = snd_device_new(card, SNDRV_DEV_LOWLEVEL, podxtpro, &ops)) < 0)
		return err;

	strcpy(card->driver, "Line6-USB");
	strcpy(card->shortname, "Line6-USB");
	sprintf(card->longname, "Line6 %s USB audio", podxtpro->devname);  /* 80 chars - see asound.h */

	if((err = snd_podxtpro_new_pcm(chip)) < 0)
		return err;

	if((err = snd_ctl_add(card, snd_ctl_new1(&podxtpro_control, chip))) < 0)
		return err;

	if((err = snd_card_register(card)) < 0)
		return err;

	spin_lock_init(&chip->lock_audio_out);
	spin_lock_init(&chip->lock_audio_in);
	spin_lock_init(&chip->lock_trigger);

	if((err = create_audio_out_urbs(chip)) < 0)
		return err;

	if((err = create_audio_in_urbs(chip)) < 0)
		return err;

	chip->volume[0] = chip->volume[1] = 128;
	return 0;
}

/* trigger callback */
int snd_podxtpro_trigger(snd_pcm_substream_t *substream, int cmd)
{
	snd_podxtpro_audio_t *chip = snd_pcm_substream_chip(substream);
	struct list_head *pos;
	snd_pcm_substream_t *s;
	int err;
	unsigned long flags;

	spin_lock_irqsave(&chip->lock_trigger, flags);

	snd_pcm_group_for_each(pos, substream) {
		s = snd_pcm_group_substream_entry(pos);

		switch(s->stream) {
		case SNDRV_PCM_STREAM_PLAYBACK:
			err = snd_podxtpro_playback_trigger(s, cmd);

			if(err < 0) {
				spin_unlock_irqrestore(&chip->lock_trigger, flags);
				return err;
			}

			break;

		case SNDRV_PCM_STREAM_CAPTURE:
			err = snd_podxtpro_capture_trigger(s, cmd);

			if(err < 0) {
				spin_unlock_irqrestore(&chip->lock_trigger, flags);
				return err;
			}

			break;

		default:
			dev_err(s2m(substream), "Unknown stream direction %d\n", s->stream);
		}
	}

	spin_unlock_irqrestore(&chip->lock_trigger, flags);
	return 0;
}

/*
	Cleanup the PODxt Pro USB audio system.
*/
void cleanup_audio(struct usb_interface *interface)
{
	int i;
	struct usb_podxtpro *podxtpro = usb_get_intfdata(interface);
	struct snd_podxtpro_audio *chip = podxtpro->chip;

	for(i = PODXTPRO_ISO_BUFFERS; i--;) {
		if(chip->urb_audio_out[i]) {
			usb_kill_urb(chip->urb_audio_out[i]);
			usb_free_urb(chip->urb_audio_out[i]);
		}
		if(chip->urb_audio_in[i]) {
			usb_kill_urb(chip->urb_audio_in[i]);
			usb_free_urb(chip->urb_audio_in[i]);
		}
	}

	if(chip->card) {
		snd_card_t *card = chip->card;
		snd_card_disconnect(card);
		snd_card_free(card);
	}

	kfree(chip);
	podxtpro->chip = 0;
}
